// standard
#include <iostream>
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <vector>

// imgui 
#include <imgui/imgui.h>
#include <imgui/imgui_impl_glut.h>
#include <imgui/imgui_impl_opengl2.h>

// free glut
#include <GL/freeglut.h>

// source
#include <vec3.h>
#include <mat4.h>
#include <model.h>

//================================
// global variables
//================================
// screen size
int g_screenWidth = 0;
int g_screenHeight = 0;

// mouse position
int mouseX = 0;
int mouseY = 0;

bool mousePressed = false;

// frame index
int g_frameIndex = 0;

// Imgui params
bool my_tool_active;
bool is_hovering = false;
float float_fov = 45;
float vec3_offset[3] = { 0, 0.3, 0 };
float vec3_cam_pos[3] = { 0, 0, 0 };
float vec3_rotation[3] = { 0, 0, 0 };
bool wireframe = false;

// model
Model my_model;
vec3 cam_pos, cam_target;

//================================
// init
//================================
void init(void) {
    // init something before main loop...

    // load model
    my_model.LoadModel("data/bunny.d");

}

//================================
// update
//================================
void update(void) {
    // do something before rendering...
}


void drawGrids(float height) {
    float step = 0.1f;
    int n = 20;
    float r = step * n;

    glBegin(GL_LINES);
    for (int i = -n; i <= n; i++) {
        glVertex3f(i * step, height, -r);
        glVertex3f(i * step, height, +r);
    }
    for (int i = -n; i <= n; i++) {
        glVertex3f(-r, height, i * step);
        glVertex3f(+r, height, i * step);
    }
    glEnd();
}

void imgui_func() {
    ImGui::Begin("Inspector", &my_tool_active, ImGuiWindowFlags_MenuBar);

    // Add your custom gui parameters here
    ImGui::Checkbox("Wire Frame", &wireframe);
    ImGui::SliderFloat("FOV", &float_fov, 40, 120);
    ImGui::DragFloat3("Obj Offset", vec3_offset, 0.1, 0.1, 0.1);
    ImGui::SliderAngle("X Axis", &vec3_rotation[0], -360, 360);
    ImGui::SliderAngle("Y Axis", &vec3_rotation[1], -360, 360);
    ImGui::SliderAngle("Z Axis", &vec3_rotation[2], -360, 360);

    ImGui::SetWindowPos(ImVec2(0, 0));
    ImGui::SetWindowSize(ImVec2(300, 200));

    is_hovering = ImGui::IsAnyWindowFocused() || ImGui::IsAnyItemFocused();
    ImGui::End();
}

//================================
// render
//================================
void render(void) {
    // Setup & Render ImGui
    ImGui_ImplOpenGL2_NewFrame();
    ImGui_ImplGLUT_NewFrame();
    imgui_func();
    ImGui::Render();

    // clear color and depth buffer
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glClearDepth(1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // enable depth test
    glEnable(GL_DEPTH_TEST);

    // Write your viewing matrix here  <------------------------------------------
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    // 1.calculate 3 camera axis
    // 2.invert the z axis (since OpenGL uses right-handed coordinate system)
    // 3.create camera transform matrix
    // 4.create camera rotation matrix
    // 5.multiply these 2 matrices
    // 6.apply the matrix under GL_MODELVIEW context
    // glLoadMatrixf( --- your viewing matrix --- );

    // draw floor
    glLineWidth(1);
    glColor4f(0.2f, 0.2f, 0.5f, 1.0f);
    drawGrids(-0.9f);

    // draw your models here  <------------------------------------------
    my_model.DrawEdges();



    // Write your projection matrix here  <------------------------------------------
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // 1.create camera projection matrix
    // 2.apply the matrix under GL_PROJECTION context
    // glLoadMatrixf( --- your projection matrix --- );




    ImGui_ImplOpenGL2_RenderDrawData(ImGui::GetDrawData());
    // swap back and front buffers
    glutSwapBuffers();
    glFlush();
}

//================================
// keyboard input
//================================
void key_press(unsigned char key, int x, int y) {
    // move your camera by changing the cam_pos from here  <------------------------------------------
    switch (key) {
    case 'w':
        break;
    case 'a':
        break;
    case 's':
        break;
    case 'd':
        break;
    default:
        break;
    }

    // ImGui callback
    ImGui_ImplGLUT_KeyboardFunc(key, x, y);
}

void special_key(int key, int x, int y) {
    switch (key) {
    case GLUT_KEY_RIGHT: //right arrow
        break;
    case GLUT_KEY_LEFT: //left arrow
        break;
    default:
        break;
    }
}


void glut_mouse_func(int button, int state, int x, int y) {
    if (state == GLUT_DOWN) {
        printf("mouse down\n");
        mouseX = x;
        mouseY = y;
        mousePressed = true;
    }
    else if (state == GLUT_UP) {
        printf("mouse up\n");
        mousePressed = false;
    }
    if ((button == 3) || (button == 4)) {
        if (state == GLUT_UP) {
            return;
        }
        printf("scrolling\n");
    }

    // ImGui callback
    ImGui_ImplGLUT_MouseFunc(button, state, x, y);
}


void glut_motion_func(int x, int y) {
    //printf("X:%d, Y:%d\n", x, y);
    if (!is_hovering) {
        mouseX = x;
        mouseY = y;
    }
    // ImGui callback
    ImGui_ImplGLUT_MotionFunc(x, y);
}

void glut_passive_motion_func(int x, int y) {
    //printf("X:%d, Y:%d\n", x, y);
    mouseX = x;
    mouseY = y;
}


//================================
// reshape : update viewport and projection matrix when the window is resized
//================================
void reshape(int w, int h) {
    // screen size
    g_screenWidth = w;
    g_screenHeight = h;

    // viewport
    glViewport(0, 0, (GLsizei)w, (GLsizei)h);

    // projection matrix <------------------------------------------
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0f, (float)w / (float)h, 0.1f, 100.0f);

    // ImGui callback
    ImGui_ImplGLUT_ReshapeFunc(g_screenWidth, g_screenHeight);
}

//================================
// timer : triggered every 16ms ( about 60 frames per second )
//================================
void timer(int value) {
    // increase frame index
    g_frameIndex++;

    update();

    // render
    glutPostRedisplay();

    // reset timer
    glutTimerFunc(16, timer, 0);
}

//================================
// main
//================================
int main(int argc, char** argv) {
    // create opengL window
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize(800, 600);
    glutInitWindowPosition(100, 100);
    glutCreateWindow(argv[0]);

    // init
    init();

    // set callback functions
    glutDisplayFunc(render);
    glutReshapeFunc(reshape);
    glutPassiveMotionFunc(glut_passive_motion_func);
    glutMotionFunc(glut_motion_func);
    glutMouseFunc(glut_mouse_func);
    glutKeyboardFunc(key_press);
    glutSpecialFunc(special_key);
    glutTimerFunc(16, timer, 0);

    // ImGui Setup
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO(); (void)io;
    ImGui::StyleColorsDark();
    ImGui::StyleColorsClassic();
    ImGui_ImplGLUT_Init();
    //ImGui_ImplGLUT_InstallFuncs();
    ImGui_ImplOpenGL2_Init();

    // main loop
    glutMainLoop();

    // ImGui Destroy
    ImGui_ImplOpenGL2_Shutdown();
    ImGui_ImplGLUT_Shutdown();
    ImGui::DestroyContext();

    return 0;
}
